/*
 * GrayStar
 * V1.0, June 2014
 * 
 * C. Ian Short
 * Saint Mary's University
 * Department of Astronomy and Physics
 * Institute for Computational Astrophysics (ICA)
 * Halifax, NS, Canada
 *  * ian.short@smu.ca
 * www.ap.smu.ca/~ishort/
 *
 * 1D, static, plane-parallel, LTE, gray stellar atmospheric model
 * core + wing approximation to Voigt spectral line profile
 *
 * Suitable for pedagogical purposes only
 * 
 * Logic written in Java SE 8.0, JDK 1.8
 * GUI written with JavaFX 8.0
 * 
 * Ported to JavaScript for deployment
 *
 * System requirements for Java version: Java run-time environment (JRE)
 * System requirements for JavaScript version: JavaScript intrepretation enabld in WWW browser (usually by default)
 *
 * Code provided "as is" - there is no formal support 
 *
 * Java default license applies:
 * End users may adapt, modify, develop, debug, and deploy at will
 * for academic and othe non-profit uses, but are asked to leave this
 * header text in place (although they may add to the header text).
 *
 */


// **********************************************

// Global variables - Doesn't work - scope not global!




var c = 2.9979249E+10; // light speed in vaccuum in cm/s
var sigma = 5.670373E-5; //Stefan-Boltzmann constant ergs/s/cm^2/K^4  
var k = 1.3806488E-16; // Boltzmann constant in ergs/K
var h = 6.62606957E-27; //Planck's constant in ergs sec
var ee = 4.80320425E-10; //fundamental charge unit in statcoulombs (cgs)
var mE = 9.10938291E-28; //electron mass (g)
//Conversion factors
var amu = 1.66053892E-24; // atomic mass unit in g
var eV = 1.602176565E-12; // eV in ergs

//Methods:
//Natural logs more useful than base 10 logs - Eg. Formal soln module: 
// Fundamental constants
var logC = Math.log(c);
var logSigma = Math.log(sigma);
var logK = Math.log(k);
var logH = Math.log(h);
var logEe = Math.log(ee); //Named so won't clash with log_10(e)
var logMe = Math.log(mE);
//Conversion factors
var logAmu = Math.log(amu);
var logEv = Math.log(eV);
// ********************************************

// *********************************************************
// 
// 
// Atmosphere astrophysical functions:
//

//log_10 Rosseland optical depth scale  
// CAUTION: Here tau[1][] is log_10!
var tauScale = function(numDeps) {

//log_10 Rosseland optical depth scale  
//Java: double tauRos[][] = new double[2][numDeps];
//var tauRos = new double[2][numDeps];
    var dummy0 = [];
    var dummy1 = [];
    dummy0.length = numDeps;
    dummy1.length = numDeps;
    var tauRos = [dummy0, dummy1];
    // Construct the log Rosseland optical depth scale:
    // Try equal spacing in log depth

    var ln10 = Math.log(10.0);
    var log10MinDepth = -6.0;
    var log10MaxDepth = 2.0;
    var logMinDepth = log10MinDepth * ln10;
    var logMaxDepth = log10MaxDepth * ln10;
    var deltaLogTau = (logMaxDepth - logMinDepth) / (numDeps - 1.0);
    var ii;
    for (var i = 0; i < numDeps; i++) {

        ii = i * 1.0;
        //Java: tauRos[1][i] = logMinDepth + ii * deltaLogTau;
        //Java: tauRos[0][i] = Math.exp(tauRos[1][i]);
        tauRos[1][i] = logMinDepth + ii * deltaLogTau;
        tauRos[0][i] = Math.exp(tauRos[1][i]);
    }

    return tauRos;
};


/**
 * Computes the Gray kinetic temperature structure, on the Rosseland optical
 * depth scale T_kin(Tau_Ros) = Teff * (0.75Tau_Ros + Hopf)^0.25
 */
var temperature = function(numDeps, teff, tauRos) {

    //Gray kinetic temeprature structure:
    // Java:double[][] temp = new double[2][numDeps];
    var dummy0 = [];
    var dummy1 = [];
    dummy0.length = numDeps;
    dummy1.length = numDeps;
    var temp = [dummy0, dummy1];
    var hopf, deltaLogTau;
    for (var i = 0; i < numDeps; i++) {

        // Interpolate approximate Hopf function:
        deltaLogTau = (tauRos[1][i] - tauRos[1][0]) / (tauRos[1][numDeps - 1] - tauRos[1][0]);
        hopf = 0.55 + deltaLogTau * (0.710 - 0.55);
        //temp[1][i] = Math.log(teff) + 
        //             0.25 * Math.log(0.75*tauRos[0][i] + 0.5);
        temp[1][i] = Math.log(teff)
                + 0.25 * Math.log(0.75 * (tauRos[0][i] + hopf));
        temp[0][i] = Math.exp(temp[1][i]);
    }

    return temp;
};

/**
 * Compute Rosseland mean extinction coefficient (cm^2/g) structure by scaling
 * from Sun
 *
 */
var kappas = function(numDeps, kappaScale, tauRos, temp, tempSun, logg, loggSun, teff) {

    // Need solar vertical Kappa_Ross structure here:
    // log(kappa_Ros) structure based on Table 9.2 of Observation
    // and Analysis of Stellar Photospheres, 3rd Ed., D.F. Gray:
    var dummy0 = [];
    var dummy1 = [];
    dummy0.length = numDeps;
    dummy1.length = numDeps;
    //var logRadius = Math.log(radius);
    var logRadiusSun = 0.0; //solar units
    var kappa = [dummy0, dummy1];
    var kappaRosSun = [dummy0, dummy1];
    var minLog10KappaRosSun = -3.5;
    var maxLog10KappaRosSun = 2.0;
    var ln10 = Math.log(10.0);
    var minLogKappaRosSun = minLog10KappaRosSun * ln10;
    var maxLogKappaRosSun = maxLog10KappaRosSun * ln10;
    var deltaKappa = (maxLogKappaRosSun - minLogKappaRosSun) / numDeps;
    var ii;
    //var fDeps = 1.0 * numDeps;
    var logHotReScale = new Array(numDeps);
    var hotReScale = new Array(numDeps);
    //Sun:
    for (var i = 0; i < numDeps; i++) {

        ii = 1.0 * i;
        kappaRosSun[1][i] = minLogKappaRosSun + ii * deltaKappa;
        kappaRosSun[0][i] = Math.exp(kappaRosSun[1][i]);
        //hotReScale[i] = (((fDeps - 1) - ii) / (fDeps - 1)) * 100.0 * (teff - 9500.0) / 9500.0;
        hotReScale[i] = ((tauRos[1][numDeps-1] - tauRos[1][i])/(tauRos[1][numDeps-1] - tauRos[1][0]) ) * 100.0 * (teff - 9500.0)/9500.0;
        logHotReScale[i] = Math.log(hotReScale[i]);
    }


    //Star:
    var logHelp, help, reScale;
    for (var i = 0; i < numDeps; i++) {

        //Try re-scaling kappaScale as a Kramer's opacity with rho proportional g
        // and T proportional to Teff

        //logHelp = (logg - loggSun) + (logRadiusSun - logRadius)
        //        + 3.5 * (tempSun[1][i] - temp[1][i]);
        //help = Math.exp(logHelp);
        //reScale = help * kappaScale;

        // No!  According to ATLAS9 ABROSS values, kappa almost independent of logg - !?
        reScale = kappaScale;
        kappa[0][i] = reScale * kappaRosSun[0][i];
        kappa[1][i] = Math.log(kappa[0][i]);

        if (teff >= 9500.0) {
            kappa[1][i] = kappa[1][i] + logHotReScale[i];
        }
        
    }

    return kappa;
};

/**
 * Solve hydrostatic eq for P scale on the tau scale - need to pick a depth
 * dependent kappa value! - dP/dTau = g/kappa --> dP/dlogTau = Tau*g/kappa 
 * 
 * press is a 4 x numDeps array: rows 0 & 1 are linear and log *gas* pressure,
 * respectively rows 2 & 3 are linear and log *radiation* pressure Split
 * pressure into gas and radiation contributions as we calculate it:
 */
var hydrostatic = function(numDeps, grav, tauRos, kappa, temp) {

    var c = 2.9979249E+10; // light speed in vaccuum in cm/s
    var sigma = 5.670373E-5; //Stefan-Boltzmann constant ergs/s/cm^2/K^4  
    var logC = Math.log(c);
    var logSigma = Math.log(sigma);
    var radFac = Math.log(4.0) + logSigma - Math.log(3.0) - logC;
    var dummy0 = [];
    var dummy1 = [];
    var dummy2 = [];
    var dummy3 = [];
    dummy0.length = numDeps;
    dummy1.length = numDeps;
    dummy2.length = numDeps;
    dummy3.length = numDeps;
    var press = [dummy0, dummy1, dummy2, dummy3];
    //double ln10 = Math.log(10.0); //handy wee quantity       
    //Upper boundary condition: total pressure at top of atmosphere
    var p1 = 1.0E-4;
    // System.out.println("HYDROSTAT: ln10= " + ln10 + " p1 " + p1 + "\r\n");
    //Finite differences in log(Tau) space - deltaX should be uniform, 
    //   but compute it on the fly anyway in case we play with other tau scales
    press[0][0] = p1;
    press[1][0] = Math.log(p1);
    press[2][0] = p1;
    press[3][0] = Math.log(p1);
    // Decalare scratch variables:
    var deltaX, deltaP, help, p2, p3;
    var logPrad, pRad, helpSub, h, k1, k2, k3, k4;
    // Calculate P at the 2nd depth point in using Euler's method:
    // deltaX = tauRos[1][1] - tauRos[1][0];
    // help = (tauRos[0][0] / kappa[0][0]) * grav;
    // deltaP = help * (deltaX);
    // p2 = p1 + deltaP;
    //  // Compute LTE bolometric radiation contribution to total HSE pressure
    // logPrad = radFac + 4.0 * temp[1][1];
    //  pRad = Math.exp(logPrad);
    // // Avoid zero or negative Pgas values in subtraction below:
    //  if (pRad >= 0.99 * p2) {
    //      pRad = 0.99 * p2;
    //  }
//
    //   // Avoid a direct subtraction in case Prad is close to Pgas for deeper 
    //  // layers of hotter stars, and both values are large:
    //  //pGas = p2 - pRad;
    //   helpSub = 1.0E0 - (pRad / p2);
    //   press[0][1] = helpSub * p2;
    //  press[1][1] = Math.log(press[0][1]);
//    press[2][1] = pRad;
    //   press[3][1] = Math.log(pRad);
    //System.out.println("HYDROSTAT: i " + i + " Pgas " + press[0][i] + " Prad " + pRad);
    //Set lower boundary of next step:
    //p1 = p2;

    // RK4 for (var i = 2; i < numDeps; i++) {  //RK4
    for (var i = 1; i < numDeps; i++) {

        // Euler's method:
        // Delta log(tau):
        deltaX = tauRos[1][i] - tauRos[1][i - 1];
        help = (tauRos[0][i] / kappa[0][i]) * grav;
        //// deltaP = help * ( deltaX ); //log10
        deltaP = help * (deltaX);
        p2 = p1 + deltaP;
        //// 4th order Runge-Kutte (mid-point), p. 705, Numerical Recipes in F77, 2nd Ed.
        // h = tauRos[1][i] - tauRos[1][i - 2];
        // k1 = h * (tauRos[0][i - 2] / kappa[0][i - 2]) * grav;
        // k2 = h * (tauRos[0][i - 1] / kappa[0][i - 1]) * grav;
        // k3 = k2;
        // k4 = h * (tauRos[0][i] / kappa[0][i]) * grav;
        //  p3 = p1 + (k1 / 6.0) + (k2 / 3.0) + (k3 / 3.0) + (k4 / 6.0);
        //System.out.println("HYDROSTAT: i " + i + " deltaX " + deltaX + 
        //                   " help " + help + " deltaP " + deltaP + " p1 " + p1 + " p2 " + p2);
        // Compute LTE bolometric radiation contribution to total HSE pressure
        logPrad = radFac + 4.0 * temp[1][i];
        pRad = Math.exp(logPrad);
        // Avoid zero or negative Pgas values in subtraction below:
        if (pRad >= 0.99 * p2) {
            pRad = 0.99 * p2;
        }  //Euler
        //  if (pRad >= 0.99 * p3) {
        //       pRad = 0.99 * p3;
        //   } // 2nd O R-K

        // Avoid a direct subtraction in case Prad is close to Pgas for deeper 
        // layers of hotter stars, and both values are large:
        //pGas = p2 - pRad;
        helpSub = 1.0E0 - (pRad / p2); //Euler
        //  helpSub = 1.0E0 - (pRad / p3); // 2nd O R-K

        press[0][i] = helpSub * p2;  //Euler
        //   press[0][i] = helpSub * p3; // 2nd O R-K
        press[1][i] = Math.log(press[0][i]);
        press[2][i] = pRad;
        press[3][i] = Math.log(pRad);
        //System.out.println("HYDROSTAT: i " + i + " Pgas " + press[0][i] + " Prad " + pRad);
        //Set lower boundary of next step:
        //p1 = p2; //Euler
        p1 = p2; // 2nd O R-K
        //   p2 = p3; // 2nd O R-K

    }

    return press;
}
;

/**
 * Solves the equation of state (EOS) for the mass density (rho) given total
 * pressure from HSE solution, for a mixture of ideal gas particles and photons
 *
 * Need to assume a mean molecular weight structure, mu(Tau)
 *
 */
var state = function(numDeps, temp, press) {

    //press is a 4 x numDeps array:
    // rows 0 & 1 are linear and log *gas* pressure, respectively
    // rows 2 & 3 are linear and log *radiation* pressure
    // double c = 9.9989E+10; // light speed in vaccuum in cm/s
    // double sigma = 5.670373E-5;   //Stefan-Boltzmann constant ergs/s/cm^2/K^4   
    //double mu = 0.62; // For now: Mean molecular weight in amu
    var mu = 1.0; // Carrol & Ostlie 2nd Ed., p. 293: mu_N = 1.3, mu_I = 0.62

    var k = 1.3806488E-16; // Boltzmann constant in ergs/K
    var amu = 1.66053892E-24; // atomic mass unit in g
    var logK = Math.log(k);
    var logAmu = Math.log(amu);
    var logMuAmu = Math.log(mu) + logAmu;
    var dummy0 = [];
    var dummy1 = [];
    dummy0.length = numDeps;
    dummy1.length = numDeps;
    var rho = [dummy0, dummy1];
    // Declare scatch variables:
    // double logPrad, pRad, pGas, logPgas;
    for (var i = 0; i < numDeps; i++) {

        // Compute LTE bolometric radiation contribution to total HSE pressure
        //logPrad = radFac + 4.0*temp[1][i] ;
        //pRad = Math.exp(logPrad);
        //pGas = press[0][i] - pRad;
        //logPgas = Math.log(pGas);
        rho[1][i] = press[1][i] - temp[1][i] + (logMuAmu - logK);
        rho[0][i] = Math.exp(rho[1][i]);
        //System.out.println("i " + i + " rho " + rho[0][i]);
    }

    return rho;
};

/** 
 * Returns vector of numDep linear geometric DEPTHS below top of atmosphere - in
 * cm (cgs) for consistency with log(g) units
 *
 * *May not be useful - scale depends on log(g)
 */
var depthScale = function(numDeps, tauRos, kappa, rho) {

    //double ln10 = Math.log(10.0); //handy wee quantity 
    //log_10 Rosseland optical depth scale  
    var depths = [];
    depths.length = numDeps;
    // Upper bounday condition: 
    // Zero point at top of atmosphere - this can be shifted later?
    // log(z) cannot really correspond to zero 
    //double logZ1 = -10.0;  // log(cm)
    //depths[0] = Math.pow(10.0, logZ1);  //cm
    var z1 = 0; //cm
    depths[0] = z1;
    var minZ = 1.0E5; // = 1km - Minimum increase in depth from one point to the next

    // declare scratch variables
    //double deltaX, deltaZ, logZ2;
    var deltaX, deltaZ, z2, z3, help, logHelp,
            h, k1, k2, k3, k4, logH, logK1, logK2, logK3, logK4;
    //Euler's method for depth at 2nd point in
    // Need to avoid using rho at upper boundary, so rho value must be taken at y_n+2 on all RHSs
    deltaX = tauRos[1][1] - tauRos[1][0];
    logHelp = tauRos[1][0] - kappa[1][0] - rho[1][2];
    //help = ( tauRos[0][0] / kappa[0][0] ) / rho[0][1];
    help = Math.exp(logHelp);
    deltaZ = (deltaX) * help;
    z2 = z1 + deltaZ;
    depths[1] = z2;
    //z1 =z2;
    for (var i = 2; i < numDeps; i++) {

        //Euler's method:
        deltaX = tauRos[1][i] - tauRos[1][i - 1];
        help = (tauRos[0][i] / kappa[0][i]) / rho[0][i];
        deltaZ = (deltaX) * help;
        z2 = z1 + deltaZ;
        depths[i] = z2;
        z1 = z2;
        /*
         
         // CAUTION; Do NOT use 4th O R-K - produces jaggedness in T(depth) relation
         
         // 4th order Runge-Kutte (mid-point), p. 705, Numerical Recipes in F77, 2nd Ed.
         h = tauRos[1][i] - tauRos[1][i-2];
         logH = Math.log(h);
         
         //k1 = h * ( tauRos[0][i-2] / kappa[0][i-2] ) / rho[0][i-1];
         //k2 = h * ( tauRos[0][i-1] / kappa[0][i-1] ) / rho[0][i-1];
         //k3 = k2;
         //4 = h * ( tauRos[0][i] / kappa[0][i] ) / rho[0][i];
         
         logK1 = logH + tauRos[1][i-2] - kappa[1][i-2] - rho[1][i];
         logK2 = logH + tauRos[1][i-1] - kappa[1][i-1] - rho[1][i];
         logK4 = logH + tauRos[1][i] - kappa[1][i] - rho[1][i];
         
         k1 = Math.exp(logK1);
         k2 = Math.exp(logK2);
         k3 = Math.exp(logK2);
         k4 = Math.exp(logK4);
         
         z3 = z1 + (k1/6.0) + (k2/3.0) + (k3/3.0) + (k4/6.0);
         if (z3 <= z2){
         z3 = z2 + minZ;
         }
         depths[i] = z3;
         z1 = z2;
         z2 = z3;
         */
    }

    return depths;
};



