/*
 * GrayStar
 * V1.0, June 2014
 * 
 * C. Ian Short
 * Saint Mary's University
 * Department of Astronomy and Physics
 * Institute for Computational Astrophysics (ICA)
 * Halifax, NS, Canada
 * ian.short@smu.ca
 * www.ap.smu.ca/~ishort/
 *
 * 1D, static, plane-parallel, LTE, gray stellar atmospheric model
 * core + wing approximation to Voigt spectral line profile
 *
 * Suitable for pedagogical purposes only
 * 
 * Logic written in Java SE 8.0, JDK 1.8
 * GUI written with JavaFX 8.0
 * 
 * Ported to JavaScript for deployment
 *
 * System requirements for Java version: Java run-time environment (JRE)
 * System requirements for JavaScript version: JavaScript intrepretation enabld in WWW browser (usually by default)
 *
 * Code provided "as is" - there is no formal support 
 *
 * Java default license applies:
 * End users may adapt, modify, develop, debug, and deploy at will
 * for academic and othe non-profit uses, but are asked to leave this
 * header text in place (although they may add to the header text).
 *
 */


// **********************************************

// Global variables - Doesn't work - scope not global!




var c = 2.9979249E+10; // light speed in vaccuum in cm/s
var sigma = 5.670373E-5; //Stefan-Boltzmann constant ergs/s/cm^2/K^4  
var k = 1.3806488E-16; // Boltzmann constant in ergs/K
var h = 6.62606957E-27; //Planck's constant in ergs sec
var ee = 4.80320425E-10; //fundamental charge unit in statcoulombs (cgs)
var mE = 9.10938291E-28; //electron mass (g)
//Conversion factors
var amu = 1.66053892E-24; // atomic mass unit in g
var eV = 1.602176565E-12; // eV in ergs

//Methods:
//Natural logs more useful than base 10 logs - Eg. Formal soln module: 
// Fundamental constants
var logC = Math.log(c);
var logSigma = Math.log(sigma);
var logK = Math.log(k);
var logH = Math.log(h);
var logEe = Math.log(ee); //Named so won't clash with log_10(e)
var logMe = Math.log(mE);
//Conversion factors
var logAmu = Math.log(amu);
var logEv = Math.log(eV);
// ********************************************

// *********************************************************
// 
// 
// Spectral line astrophysical functions:

/**
 * Line profile, phi_lambda(lambda): Assume Voigt function profile - need H(a,v)
 * Assumes CRD, LTE, ??? 
 * Input parameters: lam0 - line center wavelength in nm
 * mass - mass of absorbing particle (amu) logGammaCol - log_10(gamma) - base 10
 * logarithmic collisional (pressure) damping co-efficient (s^-1) epsilon -
 * convective microturbulence- non-thermal broadening parameter (km/s) 
 * Also needs
 * atmospheric structure information: numDeps WON'T WORK - need observer's frame
 * fixed lambda at all depths: temp structure for depth-dependent thermal line
 * broadening Teff as typical temp instead of above pressure structure, pGas, if
 * scaling gamma
 */
var lineGrid = function(lam0In, massIn, xiTIn,
        numDeps, teff) {

    var c = 2.9979249E+10; // light speed in vaccuum in cm/s
    var k = 1.3806488E-16; // Boltzmann constant in ergs/K
    var amu = 1.66053892E-24; // atomic mass unit in g
    var logC = Math.log(c);
    var logK = Math.log(k);
    var ln10 = Math.log(10.0);
    var ln2 = Math.log(2.0);
    var logE = logTen(Math.E); // for debug output

    //Put input parameters into linear cgs units:
    //double gammaCol = Math.pow(10.0, logGammaCol);
    var logTeff = Math.log(teff);
    var xiT = xiTIn * 1.0E5; //km/s to cm/s
    var lam0 = lam0In * 1.0E-7; //nm to cm
    var logLam0 = Math.log(lam0);
    var logMass = Math.log(massIn * amu); //amu to g 

    // Compute depth-independent Doppler width, Delta_lambda_D:
    var doppler, logDopp, logHelp, help;
    logHelp = ln2 + logK + logTeff - logMass; // M-B dist, square of v_mode
    help = Math.exp(logHelp) + xiT * xiT; // quadratic sum of thermal v and turbulent v
    logHelp = 0.5 * Math.log(help);
    logDopp = logHelp + logLam0 - logC;
    doppler = Math.exp(logDopp); // cm

    //Set up a half-profile Delta_lambda grid in Doppler width units 
    //from line centre to wing
    var numCore = 5;
    var numWing = 5;
    //int numWing = 0;  //debug
    var numPoints = numCore + numWing;
    // a 2D 2 X numPoints array of Delta Lambdas 
    // Row 0 : Delta lambdas in cm - will need to be in nm for Planck and Rad Trans?
    // Row 1 : Delta lambdas in Doppler widths

    var dummy1 = new Array(numPoints);
    var dummy2 = new Array(numPoints);
    var linePoints = [dummy1, dummy2];
    // Line profiel points in Doppler widths - needed for Voigt function, H(a,v):
    var v = new Array(numPoints);
    var maxCoreV = 5.0; //core half-width ~ in Doppler widths
    var maxWingDeltaLogV = 1.7 * ln10; //maximum base e logarithmic shift from line centre in Doppler widths

    var logV, ii;
    for (var il = 0; il < numPoints; il++) {

        ii = 1.0 * il;
        if (il <= numCore) {

            // In core, space v points linearly:
            // Voigt "v" parameter
            // v > 0 --> This is the *red* wing:
            v[il] = ii * maxCoreV / numCore;
            linePoints[0][il] = doppler * v[il];
            linePoints[1][il] = v[il];
        } else {

            //Space v points logarithmically in wing
            logV = ii * maxWingDeltaLogV / (numPoints - 1);
            v[il] = Math.exp(logV);
            linePoints[0][il] = doppler * v[il];
            linePoints[1][il] = v[il];
        } // end else

        //System.out.println("LineGrid: il, lam, v: " + il + " " + 
        //        linePoints[0][il] + " " + linePoints[1][il]);

    } // il lambda loop

    // Add the negative DeltaLambda half of the line:
    var numPoints2 = (2 * numPoints) - 1;
    //System.out.println("LineGrid: numpoints2: " + numPoints2);

    // Return a 2D 2 X (2xnumPoints-1) array of Delta Lambdas 
    // Row 0 : Delta lambdas in cm - will need to be in nm for Planck and Rad Trans?
    // Row 1 : Delta lambdas in Doppler widths
    var dummy3 = new Array(numPoints2);
    var dummy4 = new Array(numPoints2);
    var linePoints2 = [dummy3, dummy4];
    //wavelengths are depth-independent - just put them in the 0th depth slot:
    for (var il2 = 0; il2 < numPoints2; il2++) {

        if (il2 < numPoints - 1) {

            var il = (numPoints - 1) - il2;
            linePoints2[0][il2] = -1.0 * linePoints[0][il];
            linePoints2[1][il2] = -1.0 * linePoints[1][il];
        } else {

            //Positive DelataLambda half:   
            var il = il2 - (numPoints - 1);
            linePoints2[0][il2] = linePoints[0][il];
            linePoints2[1][il2] = linePoints[1][il];
        }

        //console.log("LineGrid: il2, lam, v: " + il2 + " " + 
        //        linePoints2[0][il2] + " " + linePoints2[1][il2]);

    } //il2 loop


    return linePoints2;
};


/**
 * Line profile, phi_lambda(lambda): Assume Voigt function profile - need H(a,v)
 * Assumes CRD, LTE, ??? 
 * Input parameters: lam0 - line center wavelength in nm
 * mass - mass of absorbing particle (amu) logGammaCol - log_10(gamma) - base 10
 * logarithmic collisional (pressure) damping co-efficient (s^-1) epsilon -
 * convective microturbulence- non-thermal broadening parameter (km/s) 
 * Also needs
 * atmospheric structure information: numDeps WON'T WORK - need observer's frame
 * fixed lambda at all depths: temp structure for depth-dependent thermal line
 * broadening Teff as typical temp instead of above pressure structure, pGas, if
 * scaling gamma
 */
var voigt = function(linePoints, lam0In, logGammaCol,
        numDeps, teff, tauRos, temp, press,
        tempSun, pressSun) {

    var c = 2.9979249E+10; // light speed in vaccuum in cm/s
    var k = 1.3806488E-16; // Boltzmann constant in ergs/K
    var logC = Math.log(c);
    var logK = Math.log(k);

    var lam0 = lam0In * 1.0E-7; //nm to cm
    var logLam0 = Math.log(lam0);
    var ln10 = Math.log(10.0);
    var ln2 = Math.log(2.0);
    var ln4pi = Math.log(4.0 * Math.PI);
    var lnSqRtPi = 0.5 * Math.log(Math.PI);
    var sqPi = Math.sqrt(Math.PI);

    var logE = logTen(Math.E); // for debug output

    var doppler = linePoints[0][1] / linePoints[1][1];
    var logDopp = Math.log(doppler);
    //System.out.println("LineProf: doppler, logDopp: " + doppler + " " + logE*logDopp);

    //Put input parameters into linear cgs units:
    //double gammaCol = Math.pow(10.0, logGammaCol);

    // Lorentzian broadening:
    // Assumes Van der Waals dominates radiative damping
    // log_10 Gamma_6 for van der Waals damping around Tau_Cont = 1 in Sun 
    //  - p. 57 of Radiative Transfer in Stellar Atmospheres (Rutten)
    var logGammaSun = 9.0 * ln10; // Convert to base e 
    //double logFudge = Math.log(2.5);  // Van der Waals enhancement factor

    var tau1 = tauPoint(numDeps, tauRos, 1.0);
    //System.out.println("LINEGRID: Tau1: " + tau1);
    //logA = 2.0 * logLam0 + logGamma - ln4pi - logC - logDopp;
    //a = Math.exp(logA);
    //System.out.println("LINEGRID: logA: " + logE * logA);
    //Set up a half-profile Delta_lambda grid in Doppler width units 
    //from line centre to wing

    var numPoints = linePoints[0].length;
    //System.out.println("LineProf: numPoints: " + numPoints);

    // Return a 2D numPoints X numDeps array of normalized line profile points (phi)

    var lineProf = new Array(numPoints);
    for (var row = 0; row < numPoints; row++) {
        lineProf[row] = new Array(numDeps);
    }
    //Java: double[][] lineProf = new double[numPoints][numDeps];

    // Line profiel points in Doppler widths - needed for Voigt function, H(a,v):
    var v = new Array(numPoints);
    var logV, ii;
//        lineProf[0][0] = 0.0; v[0] = 0.0; //Line centre - cannot do logaritmically!
    var gamma, logGamma, a, logA, voigt, core, wing, logWing, logVoigt;
    for (var id = 0; id < numDeps; id++) {

//Formula from p. 56 of Radiative Transfer in Stellar Atmospheres (Rutten),
// logarithmically with respect to solar value:
        logGamma = press[1][id] - pressSun[1][tau1] + 0.7 * (tempSun[1][tau1] - temp[1][id]) + logGammaSun;
        //logGamma = logGamma + logFudge + logGammaCol;
        logGamma = logGamma + logGammaCol;
        //System.out.println("LineGrid: logGamma: " + id + " " + logE * logGamma);

        //Voigt "a" parameter with line centre wavelength:
        logA = 2.0 * logLam0 + logGamma - ln4pi - logC - logDopp;
        a = Math.exp(logA);
        for (var il = 0; il < numPoints; il++) {

            v[il] = linePoints[1][il];
            //System.out.println("LineProf: il, v[il]: " + il + " " + v[il]);
            if (v[il] <= 2.0 && v[il] >= -2.0) {

// - Gaussian ONLY - at line centre Lorentzian will diverge!
                core = Math.exp(-1.0 * (v[il] * v[il]));
                voigt = core;
                //System.out.println("LINEGRID- CORE: core: " + core);

            } else {

                logV = Math.log(Math.abs(v[il]));
                //Gaussian core:
                core = Math.exp(-1.0 * (v[il] * v[il]));
                //System.out.println("LINEGRID- WING: core: " + core);

                logWing = logA - lnSqRtPi - (2.0 * logV);
                wing = Math.exp(logWing);
                voigt = core + wing;
                //System.out.println("LINEGRID- WING: wing: " + wing + " logV " + logV);

            } // end else

//System.out.println("LINEGRID: il, v[il]: " + il + " " + v[il] + " lineProf[0][il]: " + lineProf[0][il]);
//System.out.println("LINEGRID: il, Voigt, H(): " + il + " " + voigt);
            logVoigt = Math.log(voigt) + 2.0 * logLam0 - lnSqRtPi - logDopp - logC;
            lineProf[il][id] = Math.exp(logVoigt);
            //console.log("LineProf: il, id, lineProf[il][id]: " + il + " " + id + " " + lineProf[il][id]);
        } // il lambda loop

    } //id loop


    /*
     // Check that line profile is area-normalized, per Rutten
     // RESULT: it is NOT, area = 1.4845992503443734E-19!, but IS constant with depth - !?:
     // HOWEVER: line profile seems reasonable anyway - don't ask me.
     var delta;
     for (var id = 0; id < numDeps; id++) {
     var sum = 0.0;
     for (var il = 1; il < numPoints2; il++) {
     delta = linePoints[0][il] - linePoints[0][il - 1];
     sum = sum + (lineProf2[il][id] * delta);
     }
     System.out.println("LineGrid: id, Profile area = " + id + " " + sum );
     }
     */

    return lineProf;
};


// **************************************

// Returns depth distribution of occupation numbers in lower level of b-b transition,
// and in ground states of neutral and singly ionized stages for reference

// Input parameters:
// lam0 - line centre wavelength in nm
// logNl - log_10 column density of absorbers in lower E-level, l (cm^-2)
// logFlu - log_10 oscillator strength (unitless)
// chiL - energy of lower atomic E-level of b-b transition in eV
// chiI - ground state ionization energy to niext higher stage in (ev)
//   - we are assuming this is the neutral stage
// Also needs atsmopheric structure information:
// numDeps
// tauRos structure
// temp structure 
// rho structure
var levelPops = function(lam0In, logNlIn, ionized, chiI, chiL, linePoints, lineProf,
        numDeps, kappaScale, tauRos, temp, rho) {


    var c = 2.9979249E+10; // light speed in vaccuum in cm/s
    var k = 1.3806488E-16; // Boltzmann constant in ergs/K
    var h = 6.62606957E-27; //Planck's constant in ergs sec
    var ee = 4.80320425E-10; //fundamental charge unit in statcoulombs (cgs)
    var mE = 9.10938291E-28; //electron mass (g)
//Conversion factors
    var eV = 1.602176565E-12; // eV in ergs

//Methods:
//Natural logs more useful than base 10 logs - Eg. Formal soln module: 
// Fundamental constants
    var logC = Math.log(c);
    var logK = Math.log(k);
    var logH = Math.log(h);
    var logEe = Math.log(ee); //Named so won't clash with log_10(e)
    var logMe = Math.log(mE);
//Conversion factors
    var logEv = Math.log(eV);


    var ln10 = Math.log(10.0);
    var logE = logTen(Math.E); // for debug output
    var log2pi = Math.log(2.0 * Math.PI);
    var log2 = Math.log(2.0);

    var logNl = logNlIn * ln10;  // Convert to base e
    var logKScale = Math.log(kappaScale);

    // If we need to subtract chiI from chiL, do so *before* converting to tiny numbers in ergs!
    if (ionized) {
        chiL = chiL - chiI; 
    }

    chiI = chiI * eV;  // Convert lower E-level from eV to ergs

    var boltzFacI = chiI / k; // Pre-factor for exponent of excitation Boltzmann factor

    var logSahaFac = log2 + (3.0 / 2.0) * (log2pi + logMe + logK - 2.0 * logH);

    chiL = chiL * eV;  // Convert lower E-level from eV to ergs

    //Log of line-center wavelength in cm
    var logLam0 = Math.log(lam0In * 1.0e-7);

    // energy of b-b transition
    var logTransE = logH + logC - logLam0; //ergs
    // Energy of upper E-level of b-b transition
    var chiU = chiL + Math.exp(logTransE);  //ergs

    var boltzFacL = chiL / k; // Pre-factor for exponent of excitation Boltzmann factor
    var boltzFacU = chiU / k; // Pre-factor for exponent of excitation Boltzmann factor

    var boltzFacGround = 0.0 / k; //I know - its zero, but let's do it this way anyway'

    var refRhoIndx = tauPoint(numDeps, tauRos, 1.0);
    var refLogRho = rho[1][refRhoIndx];
    //System.out.println("LINEKAPPA: refRhoIndx, refRho " + refRhoIndx + " " + logE*refRho);

    // return a 2D 3 x numDeps array of logarithmic number densities
    // Row 0: neutral stage ground state population
    // Row 1: ionized stage ground state population
    // Row 2: level population of lower level of bb transition (could be in either stage I or II!) 
    // Row 3: level population of upper level of bb transition (could be in either stage I or II!) 
    var logNumsr0 = new Array(numDeps);
    var logNumsr1 = new Array(numDeps);
    var logNumsr2 = new Array(numDeps);
    var logNumsr3 = new Array(numDeps);
    var logNums = [logNumsr0, logNumsr1, logNumsr2, logNumsr3];

    var num, logNum, expFac, logSaha, saha, logIonFracI, logIonFracII, logNumI, logNumII;
    logNumI = 0.0;
    logNumII = 0.0;
    var logNe;

    for (var id = 0; id < numDeps; id++) {

        // reduce or enhance number density by over-all Rosseland opcity scale parameter
        logNum = logNl + logKScale;

        // scale numer density by relative depth variation of mass density
        logNum = logNum + rho[1][id] - refLogRho;

        //// reduce number density by temperature-dependent factor of Saha equation:
        // Normalize wrt to solar Teff so we don't have to evaluate all the other stuff
        // assumes constant electron number density everywhere!
        // Expression for cgs logNe for hot MS stars from *MKS* logPe expression from D. Turner (private communication):
        logNe = -4.5 - logK + 0.5 * temp[1][id] - 6.0; // last term converts m^-3 to cm^-3  

        // This assumes partition fns of unity:
        logSaha = logSahaFac - logNe - boltzFacI / temp[0][id] + (3.0) / (2.0) * temp[1][id]; // log(RHS) of standard Saha equation
        saha = Math.exp(logSaha);   //RHS of standard Saha equation
        //System.out.println("logSahaFac, logNe, logSaha= " + logE*logSahaFac + " " + logE*logNe + " " + logE*logSaha);

        logIonFracII = logSaha - Math.log(1.0 + saha); // log ionization fraction in stage II
        logIonFracI = -1.0 * Math.log(1.0 + saha);     // log ionization fraction in stage I

        // System.out.println("LevelPops: id, ionFracI, ionFracII: " + id + " " + Math.exp(logIonFracI) + " " + Math.exp(logIonFracII) );
        if (ionized) {
            //System.out.println("LevPops: ionized branch taken, ionized =  " + ionized);                

            logNums[0][id] = logNum + logIonFracI; // Ascribe entire neutral stage pop to its ground level
            logNumII = logNum + logIonFracII;
            logNums[1][id] = logNumII - boltzFacGround / temp[0][id]; // ground level of ionized stage
            logNums[2][id] = logNumII - boltzFacL / temp[0][id]; // lower level of b-b transition
            logNums[3][id] = logNumII - boltzFacU / temp[0][id]; // upper level of b-b transition

        } else {
            //System.out.println("LevPops: neutral branch taken, ionized =  " + ionized);

            logNumI = logNum + logIonFracI;
            logNums[0][id] = logNumI - boltzFacGround / temp[0][id];  // ground level of neutral stage
            logNums[1][id] = logNum + logIonFracII; // Ascribe entire ionized stage pop to its ground level
            logNums[2][id] = logNumI - boltzFacL / temp[0][id]; // lower level of b-b transition
            logNums[3][id] = logNumI - boltzFacU / temp[0][id]; // upper level of b-b transition
        }
        
            //console.log("LevelPops: id, logNums[0][id], logNums[1][id], logNums[2][id], logNums[3][id]: " + id + " "
           //        + logE * (logNums[0][id]) + " "
            //       + logE * (logNums[1][id]) + " "
            //        + logE * (logNums[2][id]) + " "
            //        + logE * (logNums[3][id]) );
      
    } //id loop

    return logNums;
};


// Assumes CRD, LTE, ???
// Input parameters:
// lam0 - line centre wavelength in nm
// logNl - log_10 column density of absorbers in lower E-level, l (cm^-2)
// logFlu - log_10 oscillator strength (unitless)
// chiL - energy of lower atomic E-level of b-b transition in eV
// chiI - ground state ionization energy to niext higher stage in (ev)
//
//   //     * PROBLEM: line kappaL values converted to mass extinction by division by rho() are 
// * not consistent with fake Kramer's Law based scaling of kappa_Ros with g.
//*   Try leaving kappaLs as linear extinctions and converting the scaled kappa_Ros back to linear units
// * with solar rho() in LineTau2
// 
// Also needs atsmopheric structure information:
// numDeps
// tauRos structure
// temp structure 
// rho structure
//var lineKap = function(lam0In, logNlIn, logFluIn, ionized, chiI, chiL, linePoints, lineProf,
//        numDeps, kappaScale, tauRos, temp, rho) {
// Level population now computed in LevelPops.levelPops():
var lineKap = function(lam0In, logNums, logFluIn, linePoints, lineProf,
        numDeps, kappaScale, tauRos, temp, rho) {

    var c = 2.9979249E+10; // light speed in vaccuum in cm/s
    var k = 1.3806488E-16; // Boltzmann constant in ergs/K
    var h = 6.62606957E-27; //Planck's constant in ergs sec
    var ee = 4.80320425E-10; //fundamental charge unit in statcoulombs (cgs)
    var mE = 9.10938291E-28; //electron mass (g)
    var eV = 1.602176565E-12; // eV in ergs
    var logC = Math.log(c);
    var logK = Math.log(k);
    var logH = Math.log(h);
    var logEe = Math.log(ee); //Named so won't clash with log_10(e)
    var logMe = Math.log(mE);

    var ln10 = Math.log(10.0);
    var logE = logTen(Math.E); // for debug output
    var log2pi = Math.log(2.0 * Math.PI);
    var log2 = Math.log(2.0);

    var lam0 = lam0In * 1.0E-7; //nm to cm
    var logLam0 = Math.log(lam0);
    //var logNl = logNlIn * ln10;  // Convert to base e
    var logFlu = logFluIn * ln10; // Convert to base e
    var logKScale = Math.log(kappaScale);

    // chiI = chiI * eV;  // Convert lower E-level from eV to ergs
    // var boltzFacI = chiI / k; // Pre-factor for exponent of excitation Boltzmann factor
//
    // var logSahaFac = log2 + (3.0 / 2.0) * (log2pi + logMe + logK - 2.0 * logH);
//
    //chiL = chiL * eV;  // Convert lower E-level from eV to ergs
    // var boltzFac = chiL / k; // Pre-factor for exponent of excitation Boltzmann factor

    var numPoints = linePoints[0].length;
    //System.out.println("LineKappa: numPoints: " + numPoints);

    var logPreFac;
    logPreFac = logFlu + Math.log(Math.PI) + 2.0 * logEe - logMe - logC;
    //System.out.println("LINEKAPPA: logPreFac " + logPreFac);

    //Assume wavelength, lambda, is constant throughout line profile for purpose
    // of computing the stimulated emission correction
    var logExpFac;
    logExpFac = logH + logC - logK - logLam0;
    //System.out.println("LINEKAPPA: logExpFac " + logExpFac);

    // var refRhoIndx = tauPoint(numDeps, tauRos, 1.0);
    // var refLogRho = rho[1][refRhoIndx];
    //System.out.println("LINEKAPPA: refRhoIndx, refRho " + refRhoIndx + " " + logE*refRho);

    // return a 2D numPoints x numDeps array of monochromatic extinction line profiles
    var logKappaL = new Array(numPoints);
    for (var row = 0; row < numPoints; row++) {
        logKappaL[row] = new Array(numDeps);
    }
    //double[][] logKappaL = new double[numPoints][numDeps];

    var num, logNum, logExpFac2, expFac, stimEm, logStimEm, logSaha, saha, logIonFrac;
    var logNe;

    for (var id = 0; id < numDeps; id++) {

        //// Level population now computed in LevelPops.levelPops()
        //// reduce or enhance number density by over-all Rosseland opcity scale parameter
        ////logNum = logNl + logKScale;
//
        //// scale numer density by relative depth variation of mass density
        ////logNum = logNum + rho[1][id] - refLogRho;
//
        //// reduce number density by temperature-dependent factor of Saha equation:
        //// Normalize wrt to solar Teff so we don't have to evaluate all the other stuff
        //// assumes constant electron number density everywhere!
//
        //// Expression for cgs logNe for hot MS stars from *MKS* logPe expression from D. Turner (private communication):
        //logNe = -4.5 - logK + 0.5 * temp[1][id] - 6.0; // last term converts m^-3 to cm^-3  
        //// This assumes partition fns of unity:
        //logSaha = logSahaFac - logNe - boltzFacI / temp[0][id];
        //saha = Math.exp(logSaha);
        ////System.out.println("logSahaFac, logNe, logSaha= " + logE*logSahaFac + " " + logE*logNe + " " + logE*logSaha);
        //if (ionized) {
        //    //System.out.println("LineKappa: ionized branch taken, ionized =  " + ionized);
        //    // If species is ionized stage, NII:
        //    logIonFrac = logSaha - Math.log(1.0 + saha);
        //} else {
        //    //System.out.println("LineKappa: neutral branch taken, ionized =  " + ionized);
        //    // If species is neutral stage, NI
        //    logIonFrac = -1.0 * Math.log(1.0 + saha);
        //}
//
        ////System.out.println("LineKappa: chiI, logIonFrac: " + chiI + " " + logE*logIonFrac);
//
        // logNum = logNum + logIonFrac;
//
        //// reduce number density by Boltzmann factor of lower E-level of b-b transition
        //// Assumes partition fn and statistical weight of unity
        //logNum = logNum - boltzFac / temp[0][id];

        logExpFac2 = logExpFac - temp[1][id];
        expFac = -1.0 * Math.exp(logExpFac2);

        stimEm = 1.0 - Math.exp(expFac);
        logStimEm = Math.log(stimEm);

        // logNums is a 2D 3 x numDeps array of logarithmic number densities
        // Row 0: neutral stage ground state population
        // Row 1: ionized stage ground state population
        // Row 2: level population of lower level of bb transition (could be in either stage I or II!) 
        // Row 3: level population of upper level of bb transition (could be in either stage I or II!) 
        logNum = logNums[2][id];

        //if (id == refRhoIndx) {
        //    System.out.println("LINEKAPPA: logStimEm " + logE*logStimEm);
        //}
        for (var il = 0; il < numPoints; il++) {

            // From Radiative Transfer in Stellar Atmospheres (Rutten), p.31
            // This is a *volume* co-efficient ("alpha_lambda") in cm^-1:
            logKappaL[il][id] = logPreFac + logStimEm + logNum + Math.log(lineProf[il][id]);
            
            //console.log("LineKap: il, id: " + il + " " + id + " logPreFac " + logE*logPreFac + " logStimEm " + logE*logStimEm + 
            //        " logNum " + logE*logNum + " lineProf " + logE * Math.log(lineProf[il][id]));

            //Convert to mass co-efficient in g/cm^2:          
            // This direct approach won't work - is not consistent with fake Kramer's law scaling of Kapp_Ros with g instead of rho
            //logKappaL[il][id] = logKappaL[il][id] - rho[1][id];
            //console.log("LineKap: il, id: " + il + " " + id + " " + " logKappaL: " + logE *logKappaL[il][id]);

        } // il - lambda loop

    } // id - depth loop

    return logKappaL;

};

/**
 * Method 2: Compute monochromatic tau scales by re-scaling from Tau_Ross -
 * PROBLEM: the approximate, scaled solar kappa_Ros values are not consistent
 * with the prescribed Tau_Ros values HOWEVER: this method removes dependence on
 * the flaky depth scale calculation and vulnerability to low rho value at the
 * surface 
 * 
 * Compute the monochromatic optical depth scale, tau_lambda, at each
 * lambda across the line profile 
 * Also computes line centre Continuum
 * monochromatic optical dpeth scale for continuum rectification And stores it
 * in array element numPoints+1 This may differ from the prescribed Tau_Ross
 * because kappa_Ross was NOT computed consistently with it 
 * 
 * * PROBLEM: line kappaL values converted to mass extinction by division by rho()
 * are not consistent with fake Kramer's Law based scaling of kappa_Ros with g.
 * Try leaving kappaLs as linear extinctions and converting the scaled kappa_Ros
 * back to linear units with solar rho()
 * 
 * Inputs: lineGrid -only neded for row 0 - wavelengths logKappaL - monochromatic line extinction
 * co-efficients kappa - we need the background continuum extinction
 * co-efficients, too!
 */
var tauLambda = function(numDeps, linePoints, logKappaL,
        kappa, tauRos, rhoSun) {

    //No monochromatic optical depth can be less than the Rosseland optical depth,
    // so prevent zero tau_lambda values by setting each tau_lambda(lambda) at the 
    //top of the atmosphere to the tau_Ross value at the top 
    // - prevents trying to take a log of zero!
    var logE = logTen(Math.E); // for debug output
    var minTauL = tauRos[0][0];
    var minLogTauL = tauRos[1][0];

    var numPoints = linePoints[0].length;

    // returns numPoints+1 x numDeps array: the numPoints+1st row holds the line centre continuum tau scale
    var logTauL = new Array(numPoints + 1);
    for (var row = 0; row < numPoints + 1; row++) {
        logTauL[row] = new Array(numDeps);
    }
    //Java: double[][] logTauL = new double[numPoints + 1][numDeps];

    var kapTot1, logKapRat1, logKapRat2, logKapRat4,
            tau1, tau2, tau3, delta, tauL,
            logInteg1, integ1, logInteg2, logInteg4,
            h, k1, k2, k3, k4, logK1, logK2, logK3, logK4,
            logKapRat, logKappaC;

    for (var il = 0; il < numPoints; il++) {

        tau1 = minTauL; //initialize accumulator
        logTauL[il][0] = minLogTauL; // Set upper boundary TauL           

        //System.out.println("LineTau: minTauL: " + minTauL);
        // Euler's method to get Tau at 2nd depth point in:
        //total extinction co-efficient

        // Convert kappa_Ros to cm^-1 for consistency with kappaL:
        logKappaC = kappa[1][0] + rhoSun[1][0];

        delta = tauRos[0][1] - tauRos[0][0];
        //logKapRat = logKappaL[il][0] - kappa[1][0];
        logKapRat = logKappaL[il][0] - logKappaC;

        tau2 = tau1 + ((Math.exp(logKapRat) + 1.0) * delta);

        logTauL[il][1] = Math.log(tau2);
        tau1 = tau2;

        for (var id = 2; id < numDeps; id++) {


            // To test: continue with Euler's method:

            // Convert kappa_Ros to cm^-1 for consistency with kappaL:
            logKappaC = kappa[1][id] + rhoSun[1][id];

            delta = tauRos[0][id] - tauRos[0][id - 1];
            //logKapRat = logKappaL[il][id] - kappa[1][id];
            logKapRat = logKappaL[il][id] - logKappaC;

            tau2 = tau1 + ((Math.exp(logKapRat) + 1.0) * delta);

            logTauL[il][id] = Math.log(tau2);
            tau1 = tau2;
            
           // console.log("tauLambda: il, id, logRhoSun, logKappaL, logKappaC, logKapRat, logTauL : " 
            //        + il + " " + id + " " + logE*rhoSun[1][id] + " " + logE*logKappaL[il][id] + " " + logE*logKappaC + " " + logE*logKapRat + " " + logE*logTauL[il][id] );

            //// 4th order Runge-Kutte (mid-point), p. 705, Numerical Recipes in F77, 2nd Ed.
            // h = tauRos[0][id] - tauRos[0][id - 2];
//
            //          logKapRat1 = logKappaL[il][id - 2] - kappa[1][id - 2];
            //        k1 = h * (Math.exp(logKapRat1) + 1.0);
            //      logKapRat2 = logKappaL[il][id - 1] - kappa[1][id - 1];
            //    k2 = h * (Math.exp(logKapRat2) + 1.0);
            //  k3 = k2;
            // logKapRat4 = logKappaL[il][id] - kappa[1][id];
//            k4 = h * (Math.exp(logKapRat4) + 1.0);
//
            //          tau3 = tau1 + (k1 / 6.0) + (k2 / 3.0) + (k3 / 3.0) + (k4 / 6.0);
//
            //          if (tau3 <= tau2) {
            //            tau3 = tau2 + minTauL;
            //      }

            //      logTauL[il][id] = Math.log(tau3);
            //    tau1 = tau2;
            //  tau2 = tau3;

        } //id loop

    } //il loop

    //This is probably superfluous here, but let's do it this way for consistency with code that was
    // dependent on Method 1:
    //Now compute the monochromatic line centre continuum optical depth scale and store it in an numPoints+1st column of
    // logTauL array:
    for (var id = 0; id < numDeps; id++) {

        logTauL[numPoints][id] = tauRos[1][id];

    }

    return logTauL;

};

/**
 * Compute the equivalent width of the Voigt line in pm - picometers NOTE: The
 * input parameter 'flux' should be a 2 x (numPoints+1) array where the
 * numPoints+1st value is the line centre monochromatic Continuum flux
 */
var eqWidth = function(flux, linePoints, lam0) {

    var Wlambda = 0.0; // Equivalent width in pm - picometers

    var numPoints = linePoints[0].length;

    var delta, term, normFlux;

    for (var il = 1; il < numPoints - 1; il++) {

        delta = linePoints[0][il] - linePoints[0][il - 1];
        delta = delta * 1.0E+7;  // cm to nm - W_lambda in pm

        // Single-point normalization to line-centre flux suitable for narrow lines:
        normFlux = flux[0][il] / flux[0][numPoints];

        term = 1.0 - normFlux;
        Wlambda = Wlambda + (term * delta);

        //System.out.println("EqWidth: il " + il + " delta " + delta + " term " + term + " normFlux " + normFlux );
        //System.out.println("EqWidth: Wlambda: " + Wlambda);
    }

    // Convert area in nm to pm - picometers
    Wlambda = Wlambda * 1.0E3;

    return Wlambda;

};



