/*
 * GrayStar
 * V1.0, June 2014
 * 
 * C. Ian Short
 * Saint Mary's University
 * Department of Astronomy and Physics
 * Institute for Computational Astrophysics (ICA)
 * Halifax, NS, Canada
 *  * ian.short@smu.ca
 * www.ap.smu.ca/~ishort/
 *
 * 1D, static, plane-parallel, LTE, gray stellar atmospheric model
 * core + wing approximation to Voigt spectral line profile
 *
 * Suitable for pedagogical purposes only
 * 
 * Logic written in Java SE 8.0, JDK 1.8
 * GUI written with JavaFX 8.0
 * 
 * Ported to JavaScript for deployment
 *
 * System requirements for Java version: Java run-time environment (JRE)
 * System requirements for JavaScript version: JavaScript intrepretation enabld in WWW browser (usually by default)
 *
 * Code provided "as is" - there is no formal support 
 *
 * Java default license applies:
 * End users may adapt, modify, develop, debug, and deploy at will
 * for academic and othe non-profit uses, but are asked to leave this
 * header text in place (although they may add to the header text).
 *
 */


// **********************************************

// Global variables - Doesn't work - scope not global!




var c = 2.9979249E+10; // light speed in vaccuum in cm/s
var sigma = 5.670373E-5; //Stefan-Boltzmann constant ergs/s/cm^2/K^4  
var k = 1.3806488E-16; // Boltzmann constant in ergs/K
var h = 6.62606957E-27; //Planck's constant in ergs sec
var ee = 4.80320425E-10; //fundamental charge unit in statcoulombs (cgs)
var mE = 9.10938291E-28; //electron mass (g)
//Conversion factors
var amu = 1.66053892E-24; // atomic mass unit in g
var eV = 1.602176565E-12; // eV in ergs

//Methods:
//Natural logs more useful than base 10 logs - Eg. Formal soln module: 
// Fundamental constants
var logC = Math.log(c);
var logSigma = Math.log(sigma);
var logK = Math.log(k);
var logH = Math.log(h);
var logEe = Math.log(ee); //Named so won't clash with log_10(e)
var logMe = Math.log(mE);
//Conversion factors
var logAmu = Math.log(amu);
var logEv = Math.log(eV);
// ********************************************
// 
// // *********************************************************
// 
// 
// Radiative transfer astrophysical functions:
//

var lamgrid = function(numLams, lamSetup) {


    var lambdaScale = [];
    lambdaScale.length = numLams;
    var logLambda;
    // Space lambdas logarithmically:
    var logLam1 = logTen(lamSetup[0]);
    var logLam2 = logTen(lamSetup[1]);
    var delta = (logLam2 - logLam1) / numLams;
    var ii;
    for (var i = 0; i < numLams; i++) {

        ii = 1.0 * i;
        logLambda = logLam1 + (ii * delta);
        lambdaScale[i] = Math.pow(10.0, logLambda);
        //System.out.println("il " + i + " lambda: " + lambdaScale[i]); //debug

    }

    return lambdaScale;
};


var thetas = function() {

    //int numMus = 10; // guess
    //double[] cosTheta = new double[numMus];
    // Try equal distribution in cos(theta) space (rather than Gaussian quadrature)

    //  cosTheta is a 2xnumMus array:
    // row 0 is used for Gaussian quadrature weights
    // row 1 is used for cos(theta) values
    // Gaussian quadrature:


    /* ***************
     "n = 21" Gaussian quadrature weights, w_i, and abscissae from 
     http://pomax.github.io/bezierinfo/legendre-gauss.html
     - ie. 11 point among 0 and positive abcissae
     
     This 11/21 of a 21-point formula: 0 plus the positive abscissae ,
     so I *think* it represents *half* the integral on the interval [-1,1],
     ie., on the interval[0,1].   SO: Divide the first weight (x=0) by 2 so 
     that the quadrature sum is exactly half the integral on [-1,1]. 
     ********** */


    //var nGauss = 11;
    //var nGauss = 7;
    var nGauss = 9;
    var theta = [];
    theta.length = nGauss;
    var weight = [];
    weight.length = nGauss;
    var cosTheta = [weight, theta];
    // I *think* the "thetas" being assigned here (abcissae) are fractional
    // angles, theta/(pi/2).

    //weight[0] = 0.5 * 0.1460811336496904;  // Divide the weight of the x=0 point by 2!  
    //weight[0] = 0.1460811336496904;
    //theta[0] = 0.0000000000000000; //disk centre
    //weight[1] = 0.1445244039899700;
    //theta[1] = 0.1455618541608951;
    //weight[2] = 0.1398873947910731;
    //theta[2] = 0.2880213168024011;
    //weight[3] = 0.1322689386333375;
    //theta[3] = 0.4243421202074388;
    //weight[4] = 0.1218314160537285;
    //theta[4] = 0.5516188358872198;
    //weight[5] = 0.1087972991671484;
    //theta[5] = 0.6671388041974123;
    //weight[6] = 0.0934444234560339;
    //theta[6] = 0.7684399634756779;
    //weight[7] = 0.0761001136283793;
    //theta[7] = 0.8533633645833173;
    //weight[8] = 0.0571344254268572;
    //theta[8] = 0.9200993341504008;
    //weight[9] = 0.0369537897708525;
    //theta[9] = 0.9672268385663063;
    //weight[10] = 0.0160172282577743;
    //theta[10] = 0.9937521706203895; //limb

    //           // 7 points on [0,1] from 13 point Gaussian quadrature on [-1,1]
    // weight[0] = 0.2325515532308739;  // disk center
    //  theta[0] = 0.0000000000000000;
    // weight[1] = 0.2262831802628972;
    //  theta[1] = 0.2304583159551348;
    //  weight[2] = 0.2078160475368885;
    //  theta[2] = 0.4484927510364469;
    //   weight[3] = 0.1781459807619457;
    //   theta[3] = 0.6423493394403402;
    //   weight[4] = 0.1388735102197872;
    //   theta[4] = 0.8015780907333099;
    //   weight[5] = 0.0921214998377285;
    //   theta[5] = 0.9175983992229779;
    //   weight[6] = 0.0404840047653159;
    //  theta[6] = 0.9841830547185881;   //limb

    // 9 points on [0,1] from 17 point Gaussian quadrature on [-1,1]    
    weight[0] = 0.1794464703562065;  //disk center
    theta[0] = 0.0000000000000000;
    weight[1] = 0.1765627053669926;
    theta[1] = 0.1784841814958479;
    weight[2] = 0.1680041021564500;
    theta[2] = 0.3512317634538763;
    weight[3] = 0.1540457610768103;
    theta[3] = 0.5126905370864769;
    weight[4] = 0.1351363684685255;
    theta[4] = 0.6576711592166907;
    weight[5] = 0.1118838471934040;
    theta[5] = 0.7815140038968014;
    weight[6] = 0.0850361483171792;
    theta[6] = 0.8802391537269859;
    weight[7] = 0.0554595293739872;
    theta[7] = 0.9506755217687678;
    weight[8] = 0.0241483028685479;
    theta[8] = 0.9905754753144174;  //limb

    for (var it = 0; it < nGauss; it++) {
        cosTheta[0][it] = weight[it];
        theta[it] = theta[it] * Math.PI / 2.0;
        cosTheta[1][it] = Math.cos(theta[it]);
    }


    return cosTheta;
};


/**
 * Formal solution of the LTE radiative transfer for the monochromatic *surface*
 * intensity, I_lambda(Tau=0, theta) at wavelength lambda
 *
 * Calls Planck.planck(lambda, temp) to get the LTE source function Input lambda
 * in nm for Planck
 */
var formalSoln = function(numDeps, cosTheta, lambda, tau, temp, lineMode) {

    var epsilon = 0.001; // tolerance for stopping deeper contriibutions to I(Tau=0)
    //  cosTheta is a 2xnumMus array:
    // row 0 is used for Gaussian quadrature weights
    // row 1 is used for cos(theta) values
    // Gaussian quadrature:
    // Number of angles, numMus, will have to be determined after the fact
    var numMus = cosTheta[0].length;
    //System.out.println("FORMALSOLN: numMus= " + numMus);
    //double[][] intens = new double[numLams][numMus];
    var intens = new Array(numMus);
    //double[] intens = new double[numMus];

    // scratch variables:
    var lnSource, lnInteg, integrand, invCosTheta, delta, newInt, increment;
    //for (int il = 0; il < numLams; il++ ) {
    for (var it = 0; it < numMus; it++) {

        invCosTheta = 1.0 / cosTheta[1][it];
        newInt = 0;
        // Extended Simpson's Rule - Numerical Recipes in F77, 2nd Ed., p. 128
        // First point in formula:
        //lnSource = planck(temp[0][0], lambda);
        //lnInteg = lnSource - (tau[0][0] * invCosTheta);
        //integrand = Math.exp(lnInteg) * invCosTheta;
        //delta = (tau[0][1] - tau[0][0]);
        //increment = (1.0 / 3.0) * integrand * delta;
        //newInt = newInt + increment;
        // for (var id = 1; id < numDeps - 1; id++) {  // Extended Simpson's Rule
        for (var id = 1; id < numDeps; id++) {   // Extended rectangle Rule

            lnSource = planck(temp[0][id], lambda);
            lnInteg = lnSource - (tau[0][id] * invCosTheta);
            integrand = Math.exp(lnInteg) * invCosTheta;
            delta = (tau[0][id] - tau[0][id - 1]);

// Extended Simpson's Rule
            //if ((id % 2) == 1) {
            //    increment = (4.0 / 3.0) * integrand * delta;
            //    newInt = newInt + increment;
            //}
//
            //          if ((id % 2) == 0) {
            //            increment = (2.0 / 3.0) * integrand * delta;
            //          newInt = newInt + increment;
            //    }

            // Extended rectangle rule:
            increment = integrand * delta;
            newInt = newInt + increment;

            // the following break-out condition is not so simple if using a closed formula: 
            // //Only keep adding contributions from deper layers if the contribution
            // // is significant

            if (tau[0][id] > 2.0 / 3.0) {
                if (newInt > 0) {
                    if (increment / newInt < epsilon) {
                        break;
                    }
                }
            }
        } //id - depth loop

        ////Last point: Extended Simpson's Rule
        //lnSource = planck(temp[0][numDeps - 1], lambda);
        //lnInteg = lnSource - (tau[0][numDeps - 1] * invCosTheta);
        //integrand = Math.exp(lnInteg) * invCosTheta;
        //delta = (tau[0][numDeps - 1] - tau[0][numDeps - 2]);
        //increment = (1.0 / 3.0) * integrand * delta;
        //newInt = newInt + increment;

        intens[it] = newInt;
    }   //it - theta loop

    //} // il - lambda loop
    return intens;
};


/**
 * Inputs: lambda: a single scalar wavelength in nm temp: a single scalar
 * temperature in K Returns log of Plank function in logBBlam - B_lambda
 * distribution in pure cgs units: ergs/s/cm^2/ster/cm
 */
var planck = function(temp, lambda) {

    //int numLams = (int) (( lamSetup[1] - lamSetup[0] ) / lamSetup[2]) + 1; 
    var logBBlam; //, BBlam;

    var c = 2.9979249E+10; // light speed in vaccuum in cm/s
    var k = 1.3806488E-16; // Boltzmann constant in ergs/K
    var h = 6.62606957E-27; //Planck's constant in ergs sec
    var logC = Math.log(c);
    var logK = Math.log(k);
    var logH = Math.log(h);
    var logPreFac = Math.log(2.0) + logH + 2.0 * logC; //log
    var logExpFac = logH + logC - logK; //log
    //double preFac = 2.0 * h * ( c * c );  //linear
    //double expFac = ( h / k ) * c;      //linear

    //System.out.println("logC " + logC + " logK " + logK + " logH " + logH);
    //System.out.println("logPreFac " + logPreFac + " logExpFac " + logExpFac);
    //Declare scratch variables:
    var logLam, logPreLamFac, logExpLamFac, expon, logExpon, denom, logDenom; //log
    //double preLamFac, expLamFac, expon, denom; //linear

    //for (int il = 0; il < numLams; il++){
    //lambda = lambda[il] * 1.0E-7;  // convert nm to cm
    lambda = lambda * 1.0E-7; // convert nm to cm
    logLam = Math.log(lambda); // Do the call to log for lambda once //log
    //System.out.println("lambda " + lambda + " logLam " + logLam);

    logPreLamFac = logPreFac - 5.0 * logLam; //log
    logExpLamFac = logExpFac - logLam; //log
    //System.out.println("logPreLamFac " + logPreLamFac + " logExpLamFac " + logExpLamFac);
    // Be VERY careful about how we divide by lambda^5:
    //preLamFac = preFac / ( lambda * lambda ); //linear
    //preLamFac = preLamFac / ( lambda * lambda );  //linear
    //preLamFac = preLamFac / lambda;   //linear
    //expLamFac = expFac / lambda;

    //for (int id = 0; id < numDeps; id++){
    //logExpon = logExpLamFac - temp[1][id];
    logExpon = logExpLamFac - Math.log(temp); //log
    //System.out.println("temp " + temp + " logTemp " + Math.log(temp));
    expon = Math.exp(logExpon); //log
    //System.out.println("logExpon " + logExpon + " expon " + expon + " denom " + denom);
    // expon = expLamFac / temp;  //linear
    denom = Math.exp(expon);
    denom = denom - 1.0;
    logDenom = Math.log(denom); //log

    //BBlam[1][id][il] = logPreLamFac - logDenom;
    //BBlam[0][id][il] = Math.exp(BBlam[1][id][il]);
    logBBlam = logPreLamFac - logDenom; //log
    // Not needed? BBlam = Math.exp(logBBlam);  //log
    //BBlam = preLamFac / denom;  //linear

    // } //id loop - depths
    // } //il loop - lambdas
    return logBBlam;
};



// For some weird reason, JavaScript does not line the function name flux()!!! - Change to flux2()

var flux2 = function(intens, cosTheta) {

    var fluxSurfSpec = new Array(2);
    // returns surface flux as a 2XnumLams vector
    //  - Row 0 linear flux (cgs units)
    //  - Row 1 log_e flux
    //  cosTheta is a 2xnumMus array:
    // row 0 is used for Gaussian quadrature weights
    // row 1 is used for cos(theta) values
    // Gaussian quadrature:
    // Number of angles, numMus, will have to be determined after the fact


    var numMus = cosTheta[0].length;
    //var numMus = 11;

    var flx = 0.0;
    for (var it = 0; it < numMus; it++) {

        flx = flx + intens[it] * cosTheta[1][it] * cosTheta[0][it];
        //flx = 1.E14;  //debug
    }  // it - theta loop

    fluxSurfSpec[0] = 2.0 * Math.PI * flx;
    fluxSurfSpec[1] = Math.log(fluxSurfSpec[0]);
    //DEBUG: fluxSurfSpec[0] = 1.0E14;
    //DEBUG: fluxSurfSpec[1] = 14.0;

    return fluxSurfSpec;
};



/**
 * First, reality check raw colours, THEN Run Vega model and subtract off Vega
 * colours for single-point calibration
 */
var UBVRI = function(lamUBVRI, lambdaScale, flux, numDeps, tauRos, temp) {



    //fluxUBVRI is a 5-element array elements:
    // [0] = U, [1] = B, [2] = V, [3] = R, [4] = I
    var colors = new Array(5);
    // Single-point calibration to Vega:
    // Veha colours computed self-consistntly with GrayFox 1.0 using 
    // Stellar parameters of Castelli (ref??):
    // Teff = 9950 K, log(g) = 3.95, ([Fe/H] = -0.5 - not directly relevent):
    //double[] vegaColors = { -0.3014,  -0.4741,  -0.3892,  -1.0016,  -0.6124 };
    var vegaColors = [-0.26780757714225084, -0.44742241728817694, -0.3522368134585257, -0.9642954846002414, -0.6120586711417157];
    //var vegaColors = [0.0, 0.0, 0.0, 0.0];
    var numUBVRI = lamUBVRI[0].length;
    var fluxUBVRILam, fluxUBVRILamr, fluxUBVRILamb,
            logFluxUBVRILam, logFluxUBVRILamr, logFluxUBVRILamb;
    var fluxUBVRI = new Array(numUBVRI);
    var fluxUBVRIr = new Array(numUBVRI);
    var fluxUBVRIb = new Array(numUBVRI);
    for (var il = 0; il < numUBVRI; il++) {

        logFluxUBVRILam = interpol2(lambdaScale, flux, lamUBVRI[0][il]);
        logFluxUBVRILamr = interpol2(lambdaScale, flux, lamUBVRI[0][il] + lamUBVRI[1][il]);
        logFluxUBVRILamb = interpol2(lambdaScale, flux, lamUBVRI[0][il] - lamUBVRI[1][il]);
        fluxUBVRI[il] = Math.exp(logFluxUBVRILam);
        fluxUBVRIr[il] = Math.exp(logFluxUBVRILamr);
        fluxUBVRIb[il] = Math.exp(logFluxUBVRILamb);
    } //il loop - lambdas

    var help, logHelp, ub, bv, vr, vi, ri,
            u, b, v, r, i, logU, logB, logV, logR, logI;
    u = (0.25 * fluxUBVRIb[0]) + (0.5 * fluxUBVRI[0]) + (0.25 * fluxUBVRIr[0]);
    logU = Math.log(u);
    b = (0.25 * fluxUBVRIb[1]) + (0.5 * fluxUBVRI[1]) + (0.25 * fluxUBVRIr[1]);
    logB = Math.log(b);
    v = (0.25 * fluxUBVRIb[2]) + (0.5 * fluxUBVRI[2]) + (0.25 * fluxUBVRIr[2]);
    logV = Math.log(v);
    r = (0.25 * fluxUBVRIb[3]) + (0.5 * fluxUBVRI[3]) + (0.25 * fluxUBVRIr[3]);
    logR = Math.log(r);
    i = (0.25 * fluxUBVRIb[4]) + (0.5 * fluxUBVRI[4]) + (0.25 * fluxUBVRIr[4]);
    logI = Math.log(i);
    // U-B: 
    //logHelp = fluxUBVRI[1][1] - fluxUBVRI[1][0];
    logHelp = logB - logU;
    help = Math.exp(logHelp);
    ub = 2.5 * logTen(help);
    colors[0] = ub - vegaColors[0];
    // B-V:
    //logHelp = fluxUBVRI[1][2] - fluxUBVRI[1][1];
    logHelp = logV - logB;
    help = Math.exp(logHelp);
    bv = 2.5 * logTen(help);
    colors[1] = bv - vegaColors[1];
    // V-R:
    //logHelp = fluxUBVRI[1][3] - fluxUBVRI[1][2];
    logHelp = logR - logV;
    help = Math.exp(logHelp);
    vr = 2.5 * logTen(help);
    colors[2] = vr - vegaColors[2];
    // V-I:
    //logHelp = fluxUBVRI[1][4] - fluxUBVRI[1][2];
    logHelp = logI - logV;
    help = Math.exp(logHelp);
    vi = 2.5 * logTen(help);
    colors[3] = vi - vegaColors[3];
    // R-I:
    //logHelp = fluxUBVRI[1][4] - fluxUBVRI[1][3];
    logHelp = logI - logR;
    help = Math.exp(logHelp);
    ri = 2.5 * logTen(help);
    colors[4] = ri - vegaColors[4];
    return colors;



}; //UBVRI



